# Job.py
import json
from collections import OrderedDict
from typing import Literal

from DIRAC.Interfaces.API.Job import Job


class MetadataDict(OrderedDict):
    MCCampaign: str
    array_layout: str
    catalogs: str
    configuration_id: str
    data_level: int
    group_size: int
    merged: int
    nsb: list
    options: str
    output_extension: str
    outputType: Literal["Data", "Log", "Model"]
    particle: str
    phiP: int
    prog_name: str
    site: str
    sct: str
    thetaP: int
    type: str
    version: str

    predefined_keys: list[str] = [
        "MCCampaign",
        "array_layout",
        "catalogs",
        "configuration_id",
        "data_level",
        "group_size",
        "merged",
        "nsb",
        "options",
        "output_extension",
        "outputType",
        "particle",
        "phiP",
        "prog_name",
        "sct",
        "site",
        "thetaP",
        "type",
        "version",
    ]

    def __setitem__(self, key, value) -> None:
        if key not in self.predefined_keys and "_prog" not in key:
            raise KeyError(f"Key '{key}' is not allowed in MetadataDict")
        super().__setitem__(key, value)


class CTAJob(Job):
    """Base Job class for CTA DL1 -> DL2 jobs"""

    def __init__(self, we_type: str) -> None:
        Job.__init__(self)
        self.we_type: str = we_type
        self.setOutputSandbox(["*Log.txt"])
        self.setName("ctajob")
        self.setTag("production")
        self.input_limit = None
        self.output_metadata = MetadataDict()
        self.output_file_metadata = MetadataDict()
        self.catalogs: str = json.dumps(["DIRACFileCatalog", "TSCatalog"])
        self.prog_name = "ctapipe-process"
        self.program_category = "analysis"
        self.software_category = "stage1"
        self.package = "ctapipe"
        self.version = "v0.10.0"
        self.compiler = "gcc48_default"
        self.configuration_id = 1
        self.data_level = 1
        self.MCCampaign = "ProdTest"
        self.options = ""
        self.data_output_pattern = "./Data/*.h5"
        self.output_type = "Data"
        self.output_data_type = "Data"
        self.output_log_type = "Log"

    def set_output_metadata(self, metadata: MetadataDict = {}) -> None:
        """Set output metadata
        Parameters:
        metadata -- metadata dictionary from telescope simulation
        """
        self.output_metadata["array_layout"] = metadata["array_layout"]
        self.output_metadata["site"] = metadata["site"]
        try:
            self.output_metadata["particle"] = metadata["particle"]
        except (KeyError, TypeError):
            pass
        try:
            phi_p = metadata["phiP"]["="]
        except (KeyError, TypeError):
            phi_p = metadata["phiP"]
        self.output_metadata["phiP"] = phi_p
        try:
            theta_p = metadata["thetaP"]["="]
        except (KeyError, TypeError):
            theta_p = metadata["thetaP"]
        self.output_metadata["thetaP"] = theta_p
        if metadata.get("sct"):
            self.output_metadata["sct"] = metadata["sct"]
        else:
            self.output_metadata["sct"] = "False"
        self.output_metadata[self.program_category + "_prog"] = self.prog_name
        self.output_metadata[self.program_category + "_prog_version"] = self.version
        self.output_metadata["data_level"] = self.data_level
        self.output_metadata["outputType"] = self.output_type
        self.output_metadata["configuration_id"] = self.configuration_id
        self.output_metadata["MCCampaign"] = self.MCCampaign

    def init_debug_step(self, i_step) -> None:
        ls_step = self.setExecutable("/bin/ls -alhtr", logFile="LS_Init_Log.txt")
        ls_step["Value"]["name"] = f"Step{i_step}_LS_Init"
        ls_step["Value"]["descr_short"] = "list files in working directory"

    def software_step(self, i_step) -> None:
        sw_step = self.setExecutable(
            "cta-prod-setup-software",
            arguments=f"-p {self.package} -v {self.version} -a {self.software_category} -g {self.compiler}",
            logFile="SetupSoftware_Log.txt",
        )
        sw_step["Value"]["name"] = f"Step{i_step}_SetupSoftware"
        sw_step["Value"]["descr_short"] = "Setup software"

    def run_dedicated_software(self, i_step) -> None:
        """To be redifined in subclasses"""
        pass

    def set_metadata_and_register_data(self, i_step) -> str:
        meta_data_json: str = json.dumps(self.output_metadata)
        file_meta_data_json: str = json.dumps(self.output_file_metadata)
        output_data_type = self.output_data_type
        dm_step = self.setExecutable(
            "cta-prod-managedata",
            arguments=f"'{meta_data_json}' '{file_meta_data_json}' {self.base_path} "
            f"'{self.data_output_pattern}' {self.package} {self.program_category} '{self.catalogs}' {output_data_type}",
            logFile="DataManagement_Log.txt",
        )
        dm_step["Value"]["name"] = f"Step{i_step}_DataManagement"
        dm_step["Value"][
            "descr_short"
        ] = "Save data files to SE and register them in DFC"
        return meta_data_json

    def mid_debug(self, i_step) -> None:
        ls_step = self.setExecutable("/bin/ls -Ralhtr", logFile="LS_End_Log.txt")
        ls_step["Value"]["name"] = f"Step{i_step}_LS_End"
        ls_step["Value"][
            "descr_short"
        ] = "list files in working directory and sub-directory"

    def register_log(self, i_step, meta_data_json) -> None:
        """To be redifined in subclasses"""
        pass

    def failover_step(self, i_step):
        failover_step = self.setExecutable(
            "/bin/ls -l", modulesList=["Script", "FailoverRequest"]
        )
        failover_step["Value"]["name"] = f"Step{i_step}_Failover"
        failover_step["Value"]["descr_short"] = "Tag files as unused if job failed"

    def end_debug_step(self, i_step):
        ls_step = self.setExecutable("/bin/ls -Ralhtr", logFile="LS_End_Log.txt")
        ls_step["Value"]["name"] = f"Step{i_step}_LSHOME_End"
        ls_step["Value"]["descr_short"] = "list files in Home directory"

    def set_executable_sequence(self, debug=False) -> None:
        i_step = 0
        if debug:
            self.init_debug_step(i_step)
            i_step += 1

        self.software_step(i_step)
        i_step += 1

        self.run_dedicated_software(i_step)
        i_step += 1

        if debug:
            self.mid_debug(i_step)
            i_step += 1

        meta_data_json: str = self.set_metadata_and_register_data(i_step)
        i_step += 1

        self.register_log(i_step, meta_data_json)
        i_step += 1

        self.failover_step(i_step)
        i_step += 1

        if debug:
            self.end_debug_step(i_step)
            i_step += 1



import json
from CTADIRAC.ProductionSystem.Client.WorkflowElement import (
    WorkflowElement,
    WorkflowElementTypes,
)
import DIRAC


def check_id(workflow_config):
    """Check step ID values"""
    for workflow_step in workflow_config:
        if not workflow_step.get("ID"):
            DIRAC.gLogger.error("Unknown step ID")
            DIRAC.exit(-1)
        elif not isinstance(workflow_step["ID"], int):
            DIRAC.gLogger.error("step ID must be integer")
            DIRAC.exit(-1)
    return True


def sort_by_id(workflow_config):
    """Sort workflow steps by ID"""
    return sorted(workflow_config, key=lambda k: k["ID"])


def check_parents(workflow_config):
    """Check if parent step is listed before child step"""
    for workflow_step in workflow_config:
        if workflow_step["input_meta_query"].get("parentID"):
            if workflow_step["input_meta_query"]["parentID"] > workflow_step["ID"]:
                DIRAC.gLogger.error(
                    "A step can only have a parent which ID is inferior to its ID"
                )
                DIRAC.exit(-1)
    return True


def instantiate_workflow_element_from_type(workflow_step, parent_prod_step):
    """Instantiate workflow element class based on the step type required"""
    wf_elt = None
    we_type: str = workflow_step["job_config"]["type"].lower()
    if we_type in WorkflowElementTypes.we_types:
        wf_elt = WorkflowElement(parent_prod_step, we_type)
    else:
        DIRAC.gLogger.error("Unknown step type")
        DIRAC.exit(-1)
    return wf_elt


def find_parent_prod_step(workflow_element_list, workflow_step):
    """Find parent prod step for a given workflow element"""
    parent_prod_step = None
    if workflow_step["input_meta_query"].get("parentID"):
        parent_prod_step = workflow_element_list[
            workflow_step["input_meta_query"]["parentID"] - 1
        ].prod_step  # Python starts indexing at 0
    return parent_prod_step


def get_parents_list(workflow_config):
    parents_list = []
    for workflow_step in workflow_config:
        if workflow_step["input_meta_query"].get("parentID"):
            parents_list.append(workflow_step["input_meta_query"]["parentID"])
    return parents_list


def check_input_source_unicity(workflow_config):
    for workflow_step in workflow_config:
        if workflow_step["input_meta_query"].get("parentID"):
            if workflow_step["input_meta_query"].get("dataset"):
                DIRAC.gLogger.error(
                    "A step cannot have input data from a dataset and from a parent"
                )
                DIRAC.exit(-1)
    return True


def check_destination_catalogs(workflow_element, workflow_step, parents_list):
    """If a change in the destination catalogs is asked by the user, check that the step does not have any children.
    This function needs to be called before setting job attributes."""
    if workflow_step["job_config"].get("catalogs"):
        if workflow_step["ID"] in parents_list:
            # Check that the default catalogs value is different from the catalogs value asked by the user
            # to issue an error
            if workflow_element.job.catalogs != json.dumps(
                workflow_step["job_config"]["catalogs"]
                .replace(", ", ",")
                .split(sep=",")
            ):
                DIRAC.gLogger.error(
                    "Catalogs can only be changed for production steps without any children."
                )
                DIRAC.exit(-1)
    return True


def check_and_sort_workflow_config(workflow_config):
    check_id(workflow_config["ProdSteps"])
    workflow_config["ProdSteps"] = sort_by_id(workflow_config["ProdSteps"])
    check_parents(workflow_config["ProdSteps"])


def instanciate_and_build_workflow_element(
    workflow_step,
    workflow_config,
    parent_prod_step,
    parents_list,
    mode,
):
    workflow_element = instantiate_workflow_element_from_type(
        workflow_step, parent_prod_step
    )
    check_destination_catalogs(workflow_element, workflow_step, parents_list)
    DIRAC.gLogger.notice(
        f"\nBuilding Production step: {workflow_step['job_config']['type']} ..."
    )
    # The order of the following instructions matters
    workflow_element.build_input_data(workflow_step)
    workflow_element.build_job_attributes(workflow_config, workflow_step)
    workflow_element.build_job_output_data(workflow_step)
    workflow_element.build_element_config(workflow_step)
    workflow_element.build_output_data()
    workflow_element.build_job_input_data(mode)
    return workflow_element



#!/bin/env python
"""
Show request given its ID, a jobID or a transformation and a task
Extension of dirac-rms-request with additional options:
--TaskStatus: selection of requests associated to a Transformation based on TaskStatus
--Stat: to print only summary informations
"""
__RCSID__ = "$Id$"

import datetime
import os

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script

Script.registerSwitch("", "Job=", "   JobID[,jobID2,...]")
Script.registerSwitch("", "Transformation=", "   transformation ID")
Script.registerSwitch(
    "", "Tasks=", "      Associated to --Transformation, list of taskIDs"
)
Script.registerSwitch("", "TaskStatus=", "      Associated to --Transformation")
Script.registerSwitch("", "Stat", "   Only print summary informations")
Script.registerSwitch("", "Verbose", "   Print more information")
Script.registerSwitch("", "Terse", "   Only print request status")
Script.registerSwitch("", "Full", "   Print full request content")
Script.registerSwitch("", "Status=", "   Select all requests in a given status")
Script.registerSwitch(
    "",
    "Since=",
    "      Associated to --Status, start date yyyy-mm-dd or nb of days (default= -one day",
)
Script.registerSwitch(
    "", "Until=", "      Associated to --Status, end date (default= now"
)
Script.registerSwitch(
    "", "Maximum=", "      Associated to --Status, max number of requests "
)
Script.registerSwitch("", "Reset", "   Reset Failed files to Waiting if any")
Script.registerSwitch("", "Force", "   Force reset even if not Failed")
Script.registerSwitch(
    "",
    "All",
    "      (if --Status Failed) all requests, otherwise exclude irrecoverable failures",
)
Script.registerSwitch("", "FixJob", "   Set job Done if the request is Done")
Script.registerSwitch("", "Cancel", "   Cancel the request")
Script.registerSwitch("", "ListJobs", " List the corresponding jobs")
Script.registerSwitch(
    "", "TargetSE=", " Select request only if that SE is in the targetSEs"
)
Script.setUsageMessage(
    "\n".join(
        [
            __doc__,
            "Usage:",
            f" {Script.scriptName} [option|cfgfile] [request[,request1,...]|<file>",
            "Arguments:",
            " request: a request ID or a unique request name",
            " <file>: a file containing a list of requests (Comma-separated on each line)",
        ]
    )
)

Script.parseCommandLine()


def convertDate(date):
    try:
        value = datetime.datetime.strptime(date, "%Y-%m-%d")
        return value
    except Exception:
        pass
    try:
        value = datetime.datetime.utcnow() - datetime.timedelta(
            hours=int(24 * float(date))
        )
    except Exception:
        gLogger.fatal("Invalid date", date)
        value = None
    return value


@Script()
def main():
    jobs = []
    requestID = 0
    transID = None
    taskIDs = None
    tasks = None
    taskStatus = None
    requests = []
    full = False
    verbose = False
    stat = False
    status = None
    until = None
    since = None
    terse = False
    allR = False
    reset = False
    fixJob = False
    maxRequests = 999999999999
    cancel = False
    listJobs = False
    force = False
    targetSE = set()
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "Job":
            try:
                jobs = []
                for arg in switch[1].split(","):
                    if os.path.exists(arg):
                        lines = open(arg).readlines()
                        jobs += [
                            int(job.strip())
                            for line in lines
                            for job in line.split(",")
                        ]
                        gLogger.notice(f"Found {len(jobs)} jobs in file {arg}")
                    else:
                        jobs.append(int(arg))
            except TypeError:
                gLogger.fatal("Invalid jobID")
        elif switch[0] == "Transformation":
            try:
                transID = int(switch[1])
            except Exception:
                gLogger.fatal("Invalid transID", switch[1])
        elif switch[0] == "Tasks":
            try:
                taskIDs = [int(task) for task in switch[1].split(",")]
            except Exception:
                gLogger.fatal("Invalid tasks", switch[1])
        elif switch[0] == "TaskStatus":
            taskStatus = switch[1].capitalize()
        elif switch[0] == "Stat":
            stat = True
        elif switch[0] == "Full":
            full = True
        elif switch[0] == "Verbose":
            verbose = True
        elif switch[0] == "Terse":
            terse = True
        elif switch[0] == "All":
            allR = True
        elif switch[0] == "Reset":
            reset = True
        elif switch[0] == "Force":
            force = True
        elif switch[0] == "Status":
            status = switch[1].capitalize()
        elif switch[0] == "Since":
            since = convertDate(switch[1])
        elif switch[0] == "Until":
            until = convertDate(switch[1])
        elif switch[0] == "FixJob":
            fixJob = True
        elif switch[0] == "Cancel":
            cancel = True
        elif switch[0] == "ListJobs":
            listJobs = True
        elif switch[0] == "Maximum":
            try:
                maxRequests = int(switch[1])
            except Exception:
                pass
        elif switch[0] == "TargetSE":
            targetSE = set(switch[1].split(","))

    if reset and not force:
        status = "Failed"
    if fixJob:
        status = "Done"
    if terse:
        verbose = True
    if status:
        if not until:
            until = datetime.datetime.utcnow()
        if not since:
            since = until - datetime.timedelta(hours=24)
    from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
    from DIRAC.RequestManagementSystem.Client.ReqClient import (
        printRequest,
        recoverableRequest,
    )
    from DIRAC.TransformationSystem.Client.TransformationClient import (
        TransformationClient,
    )

    reqClient = ReqClient()
    transClient = TransformationClient()

    if transID:
        if not taskIDs:
            condDict = {"TransformationID": transID}
            if taskStatus:
                condDict.update({"ExternalStatus": taskStatus})
            res = transClient.getTransformationTasks(condDict)
            for record in res["Value"]:
                task = record["TaskID"]
                requestID = record["ExternalID"]
                requests.append(requestID)
            if not requests:
                gLogger.error(f"No request found for transformation {transID}")
                DIRAC.exit(0)
        else:
            requests = ["%08d_%08d" % (transID, task) for task in taskIDs]
        allR = True

    elif not jobs:
        requests = []
        # Get full list of arguments, with and without comma
        for arg in [
            x.strip() for arg in Script.getPositionalArgs() for x in arg.split(",")
        ]:
            if os.path.exists(arg):
                lines = open(arg).readlines()
                requests += [
                    reqID.strip() for line in lines for reqID in line.split(",")
                ]
                gLogger.notice(f"Found {len(requests)} requests in file")
            else:
                requests.append(arg)
            allR = True
    else:
        res = reqClient.getRequestIDsForJobs(jobs)
        if not res["OK"]:
            gLogger.fatal("Error getting request for jobs", res["Message"])
            DIRAC.exit(2)
        if res["Value"]["Failed"]:
            gLogger.error(
                "No request found for jobs %s"
                % ",".join(sorted(str(job) for job in res["Value"]["Failed"]))
            )
        requests = sorted(res["Value"]["Successful"].values())
        if requests:
            allR = True
        else:
            DIRAC.exit(0)

    if status and not requests:
        allR = allR or status != "Failed"
        res = reqClient.getRequestIDsList(
            [status], limit=maxRequests, since=since, until=until
        )

        if not res["OK"]:
            gLogger.error("Error getting requests:", res["Message"])
            DIRAC.exit(2)
        requests = [
            reqID
            for reqID, _st, updTime in res["Value"]
            if updTime > since and updTime <= until and reqID
        ]
        gLogger.notice(
            f"Obtained {len(requests)} requests {status} between {since} and {until}"
        )
    if not requests:
        gLogger.notice("No request selected....")
        Script.showHelp(exitCode=2)
    okRequests = []
    warningPrinted = False
    jobIDList = []
    for reqID in requests:
        # We allow reqID to be the requestName if it is unique
        try:
            requestID = int(reqID)
        except ValueError:
            requestID = reqClient.getRequestIDForName(reqID)
            if not requestID["OK"]:
                gLogger.notice(requestID["Message"])
                continue
            requestID = requestID["Value"]

        request = reqClient.peekRequest(requestID)
        if not request["OK"]:
            gLogger.error(request["Message"])
            DIRAC.exit(-1)

        request = request["Value"]
        if not request:
            gLogger.error(f"no such request {requestID}")
            continue
        # If no operation as the targetSE, skip
        if targetSE:
            found = False
            for op in request:
                if op.TargetSE and targetSE.intersection(op.TargetSE.split(",")):
                    found = True
                    break
            if not found:
                continue
        # keep a list of jobIDs if requested
        if request.JobID and listJobs:
            jobIDList.append(request.JobID)

        if status and request.Status != status:
            gLogger.notice(
                "Request %s is not in requested status %s%s"
                % (reqID, status, " (cannot be reset)" if reset else "")
            )
            continue

        if fixJob and request.Status == "Done" and request.JobID:
            # The request is for a job and is Done, verify that the job is in the proper status
            result = reqClient.finalizeRequest(
                request.RequestID, request.JobID, useCertificates=False
            )
            if not result["OK"]:
                gLogger.error("Error finalizing job", result["Message"])
            else:
                gLogger.notice(
                    "Job %d updated to %s" % (request.JobID, result["Value"])
                )
            continue

        if cancel:
            if request.Status not in ("Done", "Failed"):
                ret = reqClient.cancelRequest(requestID)
                if not ret["OK"]:
                    gLogger.error(f"Error canceling request {reqID}", ret["Message"])
                else:
                    gLogger.notice(f"Request {reqID} cancelled")
            else:
                gLogger.notice(
                    f"Request {reqID} is in status {request.Status}, not cancelled"
                )

        elif allR or recoverableRequest(request):
            okRequests.append(str(requestID))
            if reset:
                gLogger.notice(f"============ Request {requestID} =============")
                ret = reqClient.resetFailedRequest(requestID, allR=allR)
                if not ret["OK"]:
                    gLogger.error(
                        f"Error resetting request {requestID}", ret["Message"]
                    )
            else:
                if len(requests) > 1 and not stat:
                    gLogger.notice("\n===================================")
                dbStatus = reqClient.getRequestStatus(requestID).get("Value", "Unknown")
                if not stat:
                    printRequest(
                        request,
                        status=dbStatus,
                        full=full,
                        verbose=verbose,
                        terse=terse,
                    )

    if stat:
        gLogger.notice(
            "Found %d requests with RequestID: %s"
            % (len(requests), ",".join(requestID for requestID in requests))
        )

    if listJobs:
        gLogger.notice(
            f"List of {len(jobIDList)} jobs:\n",
            ",".join(str(jobID) for jobID in jobIDList),
        )

    if status and okRequests:
        from DIRAC.Core.Utilities.List import breakListIntoChunks

        gLogger.notice(f"\nList of {len(okRequests)} selected requests:")
        for reqs in breakListIntoChunks(okRequests, 100):
            gLogger.notice(",".join(reqs))


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
  Clean an existing transformation

  Usage:
    cta-transformation-clean <transID>
"""

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Base.Script import Script
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


@Script()
def main():
    Script.parseCommandLine()
    args = Script.getPositionalArgs()
    if len(args) != 1:
        Script.showHelp()

    # get arguments
    transID = args[0]

    tc = TransformationClient()
    res = tc.cleanTransformation(transID)

    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)
    else:
        DIRAC.gLogger.notice(f"Successfully cleaned transformation {transID}")
        DIRAC.exit(0)


if __name__ == "__main__":
    main()



from collections import defaultdict
from unittest.mock import MagicMock, patch
from rich.table import Table
from CTADIRAC.Core.Utilities.return_values import s_report
from CTADIRAC.TransformationSystem.Agent.TransformationFailoverAgent import (
    TransformationFailoverAgent,
)


@patch("CTADIRAC.TransformationSystem.Agent.TransformationFailoverAgent.AgentModule")
@patch("CTADIRAC.TransformationSystem.Agent.TransformationFailoverAgent.JobDB")
@patch(
    "CTADIRAC.TransformationSystem.Agent.TransformationFailoverAgent.TransformationDB"
)
@patch.object(TransformationFailoverAgent, "get_cs_options")
def init_tfa(
    mock_am, mock_job_db, mock_trans_db, mock_cs_option
) -> TransformationFailoverAgent:
    mock_cs_option.return_value = None
    tfa = TransformationFailoverAgent()

    tfa.trans_db = mock_job_db
    tfa.job_db = mock_trans_db
    return tfa


def test_should_complete_transformation() -> None:
    tfa: TransformationFailoverAgent = init_tfa()
    tfa.get_parent_type = MagicMock(return_value=None)
    tfa.transformations_to_complete = ["Processing", "Merging"]
    trans_id = 2
    is_processed_return_values: list[bool] = [True, True]

    trans_type = "MCSimulation"
    parent_id = -1
    with patch.object(
        tfa, "is_transformation_processed", side_effect=is_processed_return_values
    ):
        result = tfa.should_complete_transformation(trans_type, trans_id, parent_id)
        assert result is False

    trans_type = "Processing"
    parent_id = 1
    tfa.get_parent_type = MagicMock(return_value=["MCSimulation"])
    with patch.object(
        tfa, "is_transformation_processed", side_effect=is_processed_return_values
    ):
        result = tfa.should_complete_transformation(trans_type, trans_id, parent_id)
        assert result is True

    trans_type = "Merging"
    parent_id = 1
    tfa.get_parent_type = MagicMock(return_value=["Processing"])
    with patch.object(
        tfa, "is_transformation_processed", side_effect=is_processed_return_values
    ):
        result = tfa.should_complete_transformation(trans_type, trans_id, parent_id)
        assert result is True

    is_processed_return_values: list[bool] = [False, True]
    trans_type = "Processing"
    parent_id = 1
    tfa.get_parent_type = MagicMock(return_value=["MCSimulation"])
    with patch.object(
        tfa, "is_transformation_processed", side_effect=is_processed_return_values
    ):
        result = tfa.should_complete_transformation(trans_type, trans_id, parent_id)
        assert result is False


def test_update_tasks_status() -> None:
    tfa: TransformationFailoverAgent = init_tfa()

    jobs_attributes = {1: {"Status": "Failed"}, 2: {"Status": "Done"}}
    tasks_infos = [
        {
            "ExternalID": 1,
            "ExternalStatus": "Failed",
            "TransformationID": 123,
            "TaskID": 1,
        },
        {
            "ExternalID": 2,
            "ExternalStatus": "Done",
            "TransformationID": 456,
            "TaskID": 2,
        },
    ]

    tfa.update_tasks_status(jobs_attributes, tasks_infos)
    # Verify that the tasks status are not changed:
    assert tfa.trans_db.setTaskStatus.call_count == 0

    tasks_infos = [
        {
            "ExternalID": 1,
            "ExternalStatus": "Submitting",
            "TransformationID": 123,
            "TaskID": 1,
        },
        {
            "ExternalID": 2,
            "ExternalStatus": "Completed",
            "TransformationID": 456,
            "TaskID": 2,
        },
    ]

    tfa.update_tasks_status(jobs_attributes, tasks_infos)
    # Verify that the tasks status are updated twice
    assert tfa.trans_db.setTaskStatus.call_count == 2


lfn_1 = "/path/to/file1.txt"
lfn_2 = "/path/to/file2.txt"


def test_get_lfns_to_assign() -> None:
    tfa: TransformationFailoverAgent = init_tfa()
    # Mock the get_transformation_files method
    tfa.get_transformation_files = MagicMock(
        return_value={
            "OK": True,
            "Value": [{"TaskID": 1, "LFN": lfn_1}, {"TaskID": 2, "LFN": lfn_2}],
        }
    )

    # Mock the update_reassign_count method
    tfa.update_reassign_count = MagicMock()

    # Define some sample data
    trans_id = 123
    failed_tasks = [1, 2]

    # Call the method being tested
    lfns_to_assign = tfa.get_lfns_to_assign(trans_id, failed_tasks)

    # Assert that the method returned the expected value
    assert lfns_to_assign == {1: lfn_1, 2: lfn_2}

    # Assert that get_transformation_files was called with the correct arguments
    tfa.get_transformation_files.assert_called_with(condDict={"TaskID": failed_tasks})

    # Assert that update_reassign_count was called for each LFN
    assert tfa.update_reassign_count.call_count == 2


def test_reassign_lfn() -> None:
    tfa: TransformationFailoverAgent = init_tfa()

    # Mock the get_tasks_by_file_status method
    tfa.trans_client.get_tasks_by_file_status = MagicMock(
        return_value={"OK": True, "Value": []}
    )

    # Mock the get_lfns_to_assign method
    tfa.get_lfns_to_assign = MagicMock(return_value={})

    # Mock the setFileStatusForTransformation method
    tfa.trans_client.setFileStatusForTransformation = MagicMock(
        return_value={"OK": True}
    )

    # Define some sample data
    trans_id = 123

    # Call the method being tested
    lfns_to_assign = tfa.reassign_lfn(trans_id)

    # Assert there is noo lfns to assign
    assert lfns_to_assign == {}

    # Assert setFileStatus is not called
    assert tfa.trans_client.setFileStatusForTransformation.call_count == 0

    tfa.trans_client.get_tasks_by_file_status = MagicMock(
        return_value={"OK": True, "Value": [1, 2]}
    )
    tfa.get_lfns_to_assign = MagicMock(return_value={1: lfn_1, 2: lfn_2})
    lfns_to_assign = tfa.reassign_lfn(trans_id)

    # Assert that the method returned the expected dictionary of LFNS
    assert lfns_to_assign == {1: lfn_1, 2: lfn_2}

    # Assert that get_tasks_by_file_status was called with the correct arguments
    tfa.trans_client.get_tasks_by_file_status.assert_called_with(
        trans_id, tfa.file_status_to_check, tfa.failed_final_task_status
    )

    # Assert that get_lfns_to_assign was called with the correct arguments
    tfa.get_lfns_to_assign.assert_called_with(trans_id, [1, 2])

    # Assert that setFileStatusForTransformation was called with the correct arguments
    tfa.trans_client.setFileStatusForTransformation.assert_called_with(
        trans_id, newLFNsStatus="Unused", lfns=[lfn_1, lfn_2], force=True
    )


def test_flush_transformation() -> None:
    tfa: TransformationFailoverAgent = init_tfa()

    # Mock getTransformation:
    tfa.trans_client.getTransformation = MagicMock(
        return_value={"OK": True, "Value": {"Status": "Active"}}
    )

    # Mock get_transformation_files:
    tfa.get_transformation_files = MagicMock(return_value={"OK": True, "Value": []})

    # Mock setTransformationParameter:
    tfa.trans_client.setTransformationParameter = MagicMock(return_value={"OK": True})

    # Mock send_mail
    tfa.send_mail = MagicMock(return_value=None)

    trans_id = 2
    trans_info_dict = {"GroupSize": 10}
    parent_id = 1
    tfa.flush_transformation(trans_id, trans_info_dict, parent_id)

    # Assert that methods are not called since transformation shoudll not be flushed
    assert tfa.get_transformation_files.call_count == 0
    assert tfa.trans_client.setTransformationParameter.call_count == 0

    tfa.trans_client.getTransformation = MagicMock(
        return_value={"OK": True, "Value": {"Status": "Completed"}}
    )
    parent_id = -1
    # Allow to flush trans but don't since there is no unused files
    tfa.get_transformation_files = MagicMock(return_value={"OK": True, "Value": []})
    tfa.flush_transformation(trans_id, trans_info_dict, parent_id)

    # assert get_trans_files is called once
    assert tfa.get_transformation_files.call_count == 1
    # Assert it is still called
    assert tfa.trans_client.setTransformationParameter.call_count == 0

    # Unused files found
    tfa.get_transformation_files = MagicMock(
        return_value={"OK": True, "Value": [lfn_1, lfn_2]}
    )
    result = tfa.flush_transformation(trans_id, trans_info_dict, parent_id)

    # assert get_trans_files is called once
    assert tfa.get_transformation_files.call_count == 1
    # Assert it is still called
    assert tfa.trans_client.setTransformationParameter.call_count == 1

    # Assert the message is return
    assert f"{trans_id} has been Flushed" in result["Value"]["Message"]


def test_parse_transformation_jobs_attributes() -> None:
    tfa: TransformationFailoverAgent = init_tfa()

    trans_id = 1
    jobs_attributes = {}

    result = tfa.parse_transformation_jobs_attributes(trans_id, jobs_attributes)
    # Assert there is an error message
    assert isinstance(result["Message"], str)

    jobs_attributes = {
        1: {
            "Status": "Done",
            "MinorStatus": "minorStatus1",
            "ApplicationStatus": "appStatus1",
            "Site": "Site1",
        },
        2: {
            "Status": "Done",
            "MinorStatus": "minorStatus2",
            "ApplicationStatus": "appStatus2",
            "Site": "Site2",
        },
    }

    result = tfa.parse_transformation_jobs_attributes(trans_id, jobs_attributes)
    # Assert there is a report
    assert isinstance(result["Value"], str)


def test_create_transformation_jobs_report() -> None:
    tfa: TransformationFailoverAgent = init_tfa()

    trans_id = 123
    attributes_of_interest = ["Status", "MinorStatus", "ApplicationStatus", "Site"]
    count_status_type = {
        ("Done", "minorStatus1", "appStatus1", "Site1"): 2,
        ("Done", "minorStatus2", "appStatus2", "Site2"): 3,
    }
    count_minor_status = {"minorStatus1": 2, "minorStatus2": 3}
    count_status = {"Done": 5}
    total_nb_jobs = 5

    # Call the method being tested
    report = tfa.create_transformation_jobs_report(
        trans_id,
        attributes_of_interest,
        count_status_type,
        count_minor_status,
        count_status,
        total_nb_jobs,
    )

    # Assert that the report is generated correctly
    assert isinstance(report, str)
    assert "Transformation: 123" in report


def test_add_rows_in_table():
    tfa: TransformationFailoverAgent = init_tfa()
    # Define mock input parameters
    count_minor_status = {"minorStatus1": 2, "minorStatus2": 3}
    count_status = {"Done": 5}
    count_status_type = {
        ("Done", "minorStatus1", "appStatus1", "Site1"): 2,
        ("Done", "minorStatus2", "appStatus2", "Site2"): 3,
    }
    total_nb_jobs = 5
    trans_id = 123
    table = Table()

    # Call the method being tested
    tfa.add_rows_in_table(
        count_minor_status,
        count_status,
        count_status_type,
        total_nb_jobs,
        trans_id,
        table,
    )

    assert table.row_count == 6


def test_get_next_status() -> None:
    tfa: TransformationFailoverAgent = init_tfa()

    # Test case 1: n < len(count_status_type_sorted)
    n = 0
    count_status_type_sorted = [
        ("Status1", "MinorStatus1"),
        ("Status2", "MinorStatus2"),
    ]
    expected_next_status = "Status2"
    expected_next_minor_status = "MinorStatus2"
    next_status, next_minor_status = tfa.get_next_status(n, count_status_type_sorted)
    assert next_status == expected_next_status
    assert next_minor_status == expected_next_minor_status

    # Test case 2: n == len(count_status_type_sorted)
    n = 1
    count_status_type_sorted = [
        ("Status1", "MinorStatus1"),
        ("Status2", "MinorStatus2"),
    ]
    expected_next_status = ""
    expected_next_minor_status = ""
    next_status, next_minor_status = tfa.get_next_status(n, count_status_type_sorted)
    assert next_status == expected_next_status
    assert next_minor_status == expected_next_minor_status


def test_create_threads_for_each_transformations():
    # Mock data
    transformations = {
        "Value": {
            "1": {"Type": "Type1"},
            "2": {"Type": "Type2"},
        }
    }

    # Mock the ThreadPoolExecutor object
    mock_thread_pool_executor = MagicMock()

    # Mock the submit method of the executor instance
    mock_future_1 = MagicMock()
    mock_future_2 = MagicMock()
    mock_thread_pool_executor.submit.side_effect = [mock_future_1, mock_future_2]

    # Create an instance of YourClass
    tfa: TransformationFailoverAgent = init_tfa()
    tfa.thread_pool_executor = mock_thread_pool_executor
    # Call the method to be tested
    future_to_trans_id = tfa.create_threads_for_each_transformations(transformations)

    # Assertions
    assert isinstance(future_to_trans_id, dict)
    assert (
        len(future_to_trans_id) == 2
    )  # Check if both futures are added to the dictionary
    assert mock_future_1 in future_to_trans_id
    assert mock_future_2 in future_to_trans_id
    assert future_to_trans_id[mock_future_1] == (
        1,
        "Type1",
    )  # Check the values associated with the future
    assert future_to_trans_id[mock_future_2] == (
        2,
        "Type2",
    )  # Check the values associated with the future


def test_check_transformation() -> None:
    trans_id = 123
    trans_info_dict = {"Type": "Active"}
    tfa: TransformationFailoverAgent = init_tfa()

    # Define mock return values for methods called within check_transformation
    tfa.get_parent_transformation_id = MagicMock(return_value=-1)
    tfa.get_transformation_tasks = MagicMock(
        return_value={"Value": [{"ExternalID": "1"}, {"ExternalID": "2"}]}
    )
    tfa.update_tasks_status = MagicMock()
    tfa.should_complete_transformation = MagicMock(return_value=False)
    tfa.transformation_to_flush = []
    tfa.transformation_to_reschedule = []
    tfa.transformation_to_report = ["Active"]
    tfa.reassign_lfn = MagicMock(return_value={})
    tfa.create_reassigned_lfn_report = MagicMock(return_value="Report")

    excpted_return = {"OK": True, "Value": "Report"}
    tfa.parse_transformation_jobs_attributes = MagicMock(return_value=excpted_return)

    expected_result = s_report(True, excpted_return["Value"])
    result = tfa.check_transformation(trans_id, trans_info_dict)
    assert result == expected_result


@patch("concurrent.futures")
def test_execute(mock_concurrent_futures):
    tfa: TransformationFailoverAgent = init_tfa()
    tfa.transformation_status = "Active"
    tfa.transformation_types = ["Type1, Type2"]
    tfa.beforeExecuting = MagicMock()
    tfa.get_eligible_transformation = MagicMock(
        return_value={
            "OK": True,
            "Value": {"trans_id1": {"Type": "Type1"}, "trans_id2": {"Type": "Type2"}},
        }
    )

    # Mock future result
    mock_future1 = MagicMock()
    mock_future2 = MagicMock()
    excpeted_return_1 = {"OK": True, "Value": "Report1"}
    excpeted_return_2 = {"OK": True, "Value": "Report2"}
    mock_future1.result.return_value = s_report(True, excpeted_return_1["Value"])
    mock_future2.result.return_value = s_report(True, excpeted_return_2["Value"])
    expected_future_result = {
        mock_future1: ("trans_id1", "Type1"),
        mock_future2: ("trans_id2", "Type2"),
    }
    tfa.create_threads_for_each_transformations = MagicMock(
        return_value=expected_future_result
    )
    mock_concurrent_futures.as_completed.return_value = [mock_future1, mock_future2]

    tfa.create_mail_message = MagicMock()
    tfa.send_mail = MagicMock()

    # Call the method being tested
    result = tfa.execute()

    expected_trans_report_result = defaultdict(dict)
    expected_trans_report_result["Type1"] = {"trans_id1": "Report1"}
    expected_trans_report_result["Type2"] = {"trans_id2": "Report2"}
    tfa.create_mail_message.assert_called_once_with(expected_trans_report_result)

    expected_result = {"OK": True, "Value": None}
    assert result == expected_result



#!/usr/bin/env python
"""
Launch a transformation from CWL workflow descriptions

Usage:
    cta-prod-submit-from-cwl-workflow <name of the Transformation> <path of the directory containing CWL files> <CWL file with workflow description> <YAML file with workflow inputs>

Examples:
    cta-prod-submit-from-cwl-workflow Transformation_test ../CWL ../CWL/simulation-run.cwl ../CWL/simulation-run.yml
"""

__RCSID__ = "$Id$"

import re
import DIRAC
from DIRAC.Core.Base.Script import Script
from CTADIRAC.Interfaces.API.MCPipeJob import MCPipeJob
from DIRAC.TransformationSystem.Client.Transformation import Transformation
from CTADIRAC.ProductionSystem.CWL.CWLWorkflow import Workflow


def submit_transformation_from_workflow(transfo, transfo_name, cwl_file, yml_file):
    """Build MC Simulation Transformation"""

    # Build Transformation
    transfo.Name = "TransformationTest"
    transfo.setTransformationName(transfo_name)  # this must be unique
    transfo.setType("MCSimulation")
    transfo.setDescription("Prod6 MC Pipe TS")
    transfo.setLongDescription("Prod6 simulation pipeline")  # mandatory

    # Build Workflow
    wf = Workflow()
    wf.load(cwl_file, yml_file)
    cmd_list = wf.run_workflow(yml_file)

    # Build Job
    MCJob = MCPipeJob()
    MCJob.setType("MCSimulation")
    MCJob.setOutputSandbox(["*Log.txt"])

    # Build steps
    i_step = 1
    for cmd in cmd_list:
        if "dirac_prod_run" in cmd:
            # dirac_prod_run is not a command, add ./ for executable
            cmd = "./" + cmd
            # Replace static run number with dynamic run number to run with DIRAC
            cmd = re.sub("--run [0-9]+", f"--run {MCJob.run_number}", cmd)

        # Run workflow
        step = MCJob.setExecutable(
            str(cmd.split(" ", 1)[0]),
            arguments=str(cmd.split(" ", 1)[1]),
            logFile="Step%i_Log.txt" % i_step,
        )
        step["Value"]["name"] = "Step%i" % i_step
        step["Value"]["descr_short"] = str(cmd.split(" ", 1)[0])
        i_step += 1

    MCJob.setExecutionEnv({"NSHOW": "10"})

    # Submit Transformation
    transfo.setBody(MCJob.workflow.toXML())
    result = transfo.addTransformation()  # transformation is created here
    if not result["OK"]:
        return result
    transfo.setStatus("Active")
    transfo.setAgentType("Automatic")
    return result


@Script()
def main():
    Script.parseCommandLine()
    argss = Script.getPositionalArgs()
    if len(argss) != 4:
        Script.showHelp()
    transfo_name = argss[0]
    config_dir = argss[1]
    cwl_file = argss[2]
    yml_file = argss[3]
    if config_dir not in cwl_file:
        Script.showHelp()
    transfo = Transformation()
    result = submit_transformation_from_workflow(
        transfo, transfo_name, cwl_file, yml_file
    )
    try:
        if not result["OK"]:
            DIRAC.gLogger.error(result["Message"])
            DIRAC.exit(-1)
        else:
            DIRAC.gLogger.notice("Done")
    except Exception:
        DIRAC.gLogger.exception()
        DIRAC.exit(-1)


if __name__ == "__main__":
    main()



from unittest.mock import patch
import pytest
import typer
from CTADIRAC.Core.Utilities.typer_callbacks import (
    job_fields_arg_callback,
    job_parameters_callback,
    job_site_callback,
    job_status_callback,
    transformation_fields_arg_callback,
    transformation_or_job_id_callback,
    transformation_status_callback,
    transformation_type_callback,
)


def test_transformation_type_callback() -> None:
    value = "MCSimulation,Processing"
    expected_result: list[str] = ["MCSimulation", "Processing"]
    result = transformation_type_callback(value)
    assert result == expected_result

    value = "InvalidType"
    with pytest.raises(typer.BadParameter):
        transformation_type_callback(value)

    value = "All"
    expected_result = None
    result: list[str] | None = transformation_type_callback(value)
    assert result == expected_result


def test_transformation_status_callback():
    value = "Active,Completed"
    expected_result: list[str] = ["Active", "Completed"]
    result = transformation_status_callback(value)
    assert result == expected_result

    value = "InvalidStatus"
    with pytest.raises(typer.BadParameter):
        transformation_status_callback(value)

    value = "All"
    expected_result = None
    result: list[str] | None = transformation_status_callback(value)
    assert result == expected_result


def test_transformation_or_job_id_callback() -> None:
    value = "123,456"
    expected_result = ["123", "456"]
    result = transformation_or_job_id_callback(value)
    assert result == expected_result

    value = "InvalidID"
    with pytest.raises(typer.BadParameter):
        transformation_or_job_id_callback(value)


def test_job_parameters_callback() -> None:
    value = "AgentLocalSE"
    result = job_parameters_callback(value)
    assert result == value

    value = "InvalidParam"
    with pytest.raises(typer.BadParameter):
        job_parameters_callback(value)


@patch("CTADIRAC.Core.Utilities.typer_callbacks.getSites")
def test_job_site_callback(mock_get_sites) -> None:
    site = "ARC.CSCS.ch"
    mock_get_sites.return_value = {"OK": True, "Value": [site]}
    value = site
    expected_result = [value]
    result = job_site_callback(value)
    assert result == expected_result

    value = "InvalidSite"
    with pytest.raises(typer.BadParameter):
        job_site_callback(value)


def test_job_status_callback() -> None:
    value = "Done,Failed"
    expected_result = ["Done", "Failed"]
    result = job_status_callback(value)
    assert result == expected_result

    value = "InvalidStatus"
    with pytest.raises(typer.BadParameter):
        job_status_callback(value)


def test_transformation_fields_arg_callback() -> None:
    value = ["TransformationID", "123,456", "Status", "Active"]
    expected_result = [{"Status": ["Active"], "TransformationID": ["123", "456"]}]
    result = transformation_fields_arg_callback(value)
    assert result == expected_result

    value = ["InvalidField", "value"]
    with pytest.raises(typer.BadParameter):
        transformation_fields_arg_callback(value)


def test_job_fields_arg_callback() -> None:
    value = ["JobGroup", "123,456", "Status", "Done,Failed"]
    expected_result = [{"Status": ["Done", "Failed"], "JobGroup": ["123", "456"]}]
    result = job_fields_arg_callback(value)
    assert result == expected_result

    value = ["InvalidField", "value"]
    with pytest.raises(typer.BadParameter):
        job_fields_arg_callback(value)



#!/usr/bin/env python
"""
Launch a production from a workflow description

Usage:
    cta-prod-submit <name of the production> <YAML file with workflow description> <execution mode>
Arguments:
   name of the production
   YAML file with the workflow description
   execution mode is optional:
     ps: submit the workflow to the Production System (default)
     wms: submit individual jobs to the WMS
     local: execute the workflow locally as individual jobs
     dry-run: get the list of files that will be processed by each transformation, without running the workflow
Example:
    cta-prod-submit TestProd production_config.yml
Example for local execution (for testing):
    cta-prod-submit TestProd production_config.yml local
"""

__RCSID__ = "$Id$"

from DIRAC.Core.Base.Script import Script

Script.parseCommandLine()

from CTADIRAC.ProductionSystem.Client.Utilities.production_utils import (
    check_and_sort_workflow_config,
    find_parent_prod_step,
    get_parents_list,
    instanciate_and_build_workflow_element,
)

import DIRAC
from ruamel.yaml import YAML

yaml = YAML(typ="safe", pure=True)
import json

from DIRAC.ProductionSystem.Client.ProductionClient import ProductionClient
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from CTADIRAC.Core.Utilities.tool_box import read_inputs_from_file


def dry_run(metaquery):
    fc = FileCatalog()
    result = fc.findFilesByMetadata(dict(metaquery))
    return result


def perform_dry_run(workflow_element, workflow_step, prod_name):
    if workflow_element.prod_step.Inputquery:
        res = dry_run(workflow_element.prod_step.Inputquery)
        if res["OK"]:
            print_dry_run_results(res, workflow_element, workflow_step, prod_name)
        else:
            DIRAC.gLogger.notice(
                f"\tStep {workflow_step['ID']} will not use existing files."
            )
    else:
        DIRAC.gLogger.notice(
            f"\tStep {workflow_step['ID']} will not process any file (no input query defined)."
        )


def print_dry_run_results(res, workflow_element, workflow_step, prod_name):
    if res["Value"]:
        files_list = res["Value"]
        print_files_to_process(files_list, workflow_element, workflow_step)
        filename = f"{prod_name}_Step{workflow_step['ID']}_input_files.list"
        with open(filename, "w") as ascii:
            for file in files_list:
                ascii.write(f"{file}\n")

        DIRAC.gLogger.notice(
            f"\nFull list of {len(files_list)} files dumped into {filename}"
        )

    else:
        DIRAC.gLogger.notice(
            f"\tStep {workflow_step['ID']} will not use existing files."
        )


def print_files_to_process(files_list, workflow_element, workflow_step):
    DIRAC.gLogger.notice(
        f"\tStep {workflow_step['ID']} will process files matching the following query :"
        f" \n{workflow_element.prod_step.Inputquery}"
    )
    if len(files_list) < 10:
        DIRAC.gLogger.notice(
            f"List of {len(files_list)} files to be processed by step {workflow_step['ID']}:"
        )
        for file in files_list:
            DIRAC.gLogger.notice(file)
    else:
        DIRAC.gLogger.notice(
            f"Example of 10 files to be processed by step {workflow_step['ID']}:"
        )
        for file in files_list[0:10]:
            DIRAC.gLogger.notice(file)


def submit_we_job(workflow_element, mode):
    res = Dirac().submitJob(workflow_element.job, mode=mode)
    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)
    DIRAC.gLogger.notice("\tSubmitted job:", res["Value"])


def build_workflow(workflow_config, prod_sys_client, prod_name, mode):
    """For each workflow step, build the associated workflow element composed of a job and a production step"""
    workflow_element_list = []
    check_and_sort_workflow_config(workflow_config)
    parents_list = get_parents_list(workflow_config["ProdSteps"])
    for workflow_step in workflow_config["ProdSteps"]:
        parent_prod_step = find_parent_prod_step(workflow_element_list, workflow_step)
        workflow_element = instanciate_and_build_workflow_element(
            workflow_step, workflow_config, parent_prod_step, parents_list, mode
        )
        prod_sys_client.addProductionStep(workflow_element.prod_step)
        workflow_element_list.append(workflow_element)

        # For dry-run mode : print the list of input files for each transformation
        if mode.lower() == "dry-run":
            perform_dry_run(workflow_element, workflow_step, prod_name)

        # For local and wms mode : build and submit the job in the same loop
        if mode.lower() in ["wms", "local"]:
            submit_we_job(workflow_element, mode)


def start_production(prod_sys_client, prod_name):
    # Get the production description
    prod_description = prod_sys_client.prodDescription
    # Create the production
    res = prod_sys_client.addProduction(prod_name, json.dumps(prod_description))
    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)

    # Start the production, i.e. instantiate the transformation steps
    res = prod_sys_client.startProduction(prod_name)

    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)
    DIRAC.gLogger.notice(f"\nProduction {prod_name} successfully created")


def print_submitted_transformations(prod_name, trans_list, trans_client):
    # Print the submitted transformations
    if not trans_list:
        DIRAC.gLogger.notice(
            f"No transformation associated with production {prod_name}"
        )
        DIRAC.exit(-1)
    for trans in trans_list:
        trans_id = trans["TransformationID"]
        trans_name = trans_client.getTransformationParameters(
            trans_id, "TransformationName"
        )["Value"]
        DIRAC.gLogger.notice(
            f"\tSubmitted transformation: {trans_name} with transID {trans_id}"
        )


def attach_files_to_transformation(job_config, trans_client, trans_list):
    trans0 = trans_list[0]["TransformationID"]
    DIRAC.gLogger.notice(
        "Using %d input files from test_input_data.list" % (job_config["input_limit"])
    )
    infile_list = read_inputs_from_file("test_input_data.list")
    res = trans_client.addFilesToTransformation(trans0, infile_list)  # Files added here
    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)
    else:
        DIRAC.gLogger.notice(
            "Successfully added %d files to transformation %s"
            % (job_config["input_limit"], trans0)
        )
        DIRAC.exit(0)


@Script()
def main():
    arguments = Script.getPositionalArgs()
    if len(arguments) not in list(range(2, 4)):
        Script.showHelp()

    # Read the arguments
    prod_name = arguments[0]
    workflow_config_file = arguments[1]
    mode = "ps"
    if len(arguments) == 3:
        mode = arguments[2]
    if mode not in ["ps", "wms", "local", "dry-run"]:
        Script.showHelp()

    with open(workflow_config_file) as stream:
        workflow_config = yaml.load(stream)

    ##################################
    # Create the production
    DIRAC.gLogger.notice(f"\nBuilding new production: {prod_name}")
    prod_sys_client = ProductionClient()
    trans_client = TransformationClient()

    ##################################
    # Build production steps according to the workflow description
    build_workflow(workflow_config, prod_sys_client, prod_name, mode)
    ##################################

    # The default mode is ps, i.e. submit the worflow to the Production System
    if mode.lower() == "ps":
        start_production(prod_sys_client, prod_name)
        res = prod_sys_client.getProductionTransformations(prod_name)
        if res["OK"]:
            trans_list = res["Value"]
            print_submitted_transformations(prod_name, trans_list, trans_client)
        else:
            DIRAC.gLogger.error(res["Message"])
            DIRAC.exit(-1)

        # If input_limit is set attach testing files to the first transformation
        prod_step_0 = workflow_config["ProdSteps"][0]
        job_config = prod_step_0["job_config"]
        if "input_limit" in job_config:
            attach_files_to_transformation(job_config, trans_client, trans_list)


########################################################
if __name__ == "__main__":
    main()



""" Collection of classes for DMS testing purposes
"""

__RCSID__ = "$Id$"

import os

from DIRAC.Core.Base.Script import Script

Script.parseCommandLine()

import DIRAC
from DIRAC import gLogger
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.Resources.Catalog.RucioFileCatalogClient import RucioFileCatalogClient
from DIRAC.DataManagementSystem.Client.DataManager import DataManager


class ClientFileCatalog:
    def __init__(self, catalog):
        self.catalogs = [catalog]
        self.dest_se_1 = "PIC-Disk"
        self.dest_se_2 = "CSCS-Disk"
        self.test_dir = "/vo.cta.in2p3.fr/tests/file_catalog"
        self.lfn = os.path.join(self.test_dir, "testFile.txt")
        self.file_name = os.path.basename(self.lfn)
        self.file_tuple = (
            self.lfn,
            "destUrl",
            0,
            self.dest_se_1,
            "D41D8CD9-8F00-B204-E980-0998ECF8427E",
            "001",
        )
        self.file_dict = dict()
        self.file_dict[self.lfn] = {
            "Checksum": "f1df822d",
            "GUID": "7FF3F2EB-1B21-4874-493E-25C27FF3F53F",
            "Size": 3303708045,
            "SE": self.dest_se_1,
            "PFN": os.path.join(
                "root://tmpes28pstr@srm.pic.es//pnfs/pic.es/data/cta/Grid/Data",
                self.lfn,
            ),
        }

        if "RucioFileCatalog" in self.catalogs:
            self.fc = RucioFileCatalogClient()
        elif "DIRACFileCatalog" in self.catalogs:
            self.fc = FileCatalogClient()
        self.dm = DataManager(self.catalogs)

    def write_test_file(self):
        with open(self.file_name, "w") as file:
            file.write("This is a test file")
        return DIRAC.S_OK()

    def create_directory(self):
        gLogger.notice(f"Create directory {self.test_dir}")
        res = self.fc.createDirectory(self.test_dir)
        if not res["OK"]:
            gLogger.error(f"ERROR: Failed to create directory: {res['Message']}")
        return res

    def list_directory(self):
        gLogger.notice(f"List directory {self.test_dir}")
        res = self.fc.listDirectory(self.test_dir)
        if not res["OK"]:
            gLogger.error(f"ERROR: Failed to list directory: {res['Message']}")
        return res

    def register_file(self):
        "Only registers file in the catalog without uploading to any SE"
        gLogger.notice(f"Register file {self.lfn}")
        res = self.fc.addFile(self.file_dict)
        if not res["OK"]:
            gLogger.error(f"ERROR: Failed to register file: {res['Message']}")
        return res

    def put_and_register(self):
        "Use the DataManager for the put and register"
        gLogger.notice(f"Upload file {self.lfn} to {self.dest_se_1}")
        res = self.dm.putAndRegister(self.lfn, self.file_name, self.dest_se_1)
        if not res["OK"]:
            gLogger.error(f"ERROR: Failed to upload file: {res['Message']}")
        return res

    def remove_file(self):
        gLogger.notice(f"Remove file {self.lfn}")
        res = self.fc.removeFile(self.lfn)
        if not res["OK"]:
            gLogger.error(f"ERROR: Failed to remove file: {res['Message']}")
        return res



#!/usr/bin/env python
""" Collection of simple functions to verify each of
    the main Prod3MCJob steps
"""

# generic imports
import os
import glob

# DIRAC imports
import DIRAC
from DIRAC.Core.Base.Script import Script

Script.setUsageMessage(
    "\n".join(
        [
            __doc__.split("\n")[1],
            "Usage:",
            f"  {Script.scriptName} stepName",
            "Arguments:",
            "  stepName: corsika, simtel, merging",
            f"\ne.g: {Script.scriptName} corsika",
        ]
    )
)

Script.parseCommandLine()


def clean_output_file(output_files):
    """Delete Local Files"""
    DIRAC.gLogger.notice("Deleting Local Files")
    for afile in output_files:
        DIRAC.gLogger.warn(f"Remove local File {afile.strip()}")
        os.remove(afile.strip())


def verify_corsika():
    """Verify a generic Corsika log file"""
    DIRAC.gLogger.notice("Verifying Corsika log file")

    # get list of output files
    log_file = glob.glob("Data/corsika/run*/run*.log")
    if len(log_file) != 1:
        DIRAC.gLogger.error("no log file found!")
        DIRAC.exit(-1)

    # check EOR tag
    tag = "=== END OF RUN ==="
    content = open(log_file[0]).read()
    if content.find(tag) < 0:
        DIRAC.gLogger.error(f'"{tag}" tag not found!')
        corsika_files = list(os.popen('find . -iname "*corsika.*z*"'))
        if len(corsika_files) > 0:
            clean_output_file(corsika_files)
        simtel_files = list(os.popen('find . -iname "*simtel.*z*"'))
        if len(simtel_files) > 0:
            clean_output_file(simtel_files)
        log_gz_files = list(os.popen('find . -iname "*log.*z*"'))
        if len(log_gz_files) > 0:
            clean_output_file(log_gz_files)
        log_hist_files = list(os.popen('find . -iname "*log_hist.tar"'))
        if len(log_hist_files) > 0:
            clean_output_file(log_hist_files)
        DIRAC.exit(-1)


def verifySimtel(nbFiles=31, minSize=50.0):
    """Verify a PROD3 simtel step

    Keyword arguments:
    nbFiles -- number of output files expected
    minSize -- minimum file size
    """
    DIRAC.gLogger.notice("Verifying Simtel step")
    # get list of output files
    outputFiles = glob.glob("Data/simtel_tmp/Data/*simtel.*z*")

    # check the number of output files --- could be done by telescope type
    N = len(outputFiles)
    if N != nbFiles:
        DIRAC.gLogger.error(f"Wrong number of Simtel files : {N} instead of {nbFiles}")
        clean_output_file(outputFiles)
        DIRAC.exit(-1)

    # check the file size --- could be done by telescope type
    for afile in outputFiles:
        sizekb = os.path.getsize(afile) / 1024.0
        if sizekb < minSize:
            DIRAC.gLogger.error(
                f"{afile}\n File size too small : {sizekb} < {minSize} kb"
            )
            clean_output_file(outputFiles)
            DIRAC.exit(-1)
    DIRAC.gLogger.notice("Good files found:\n%s" % "\n".join(outputFiles))


def verifyMerging(nbFiles=10, minSize=5000.0):
    """Verify a PROD3 simtel merging step

    Keyword arguments:
    nbFiles -- number of output files expected
    minSize -- minimum file size
    """
    DIRAC.gLogger.notice("Verifying Merging step")

    # get list of output files
    outputFiles = glob.glob("Data/sim_telarray/cta-prod3/0.0deg/Data/*simtel.*z*")

    # check the number of output files --- could be done by telescope type
    N = len(outputFiles)
    if N != nbFiles:
        DIRAC.gLogger.error(
            f"Wrong number of Simtel Merged files : {N} instead of {nbFiles}"
        )
        clean_output_file(outputFiles)
        DIRAC.exit(-1)

    # check the file size --- could be done by telescope type
    for afile in outputFiles:
        sizekb = os.path.getsize(afile) / 1024.0
        if sizekb < minSize:
            DIRAC.gLogger.error(
                f"{afile}\n File size too small : {sizekb} < {minSize} kb"
            )
            clean_output_file(outputFiles)
            DIRAC.exit(-1)
    DIRAC.gLogger.notice("Good files found:\n%s" % "\n".join(outputFiles))


def verifyAnalysisInputs(minSize=50.0):
    """Verify input files for analysis

    Keyword arguments:
    minSize -- minimum file size
    """
    DIRAC.gLogger.notice("Verifying AnalysisInputs step")

    # get list of output files
    outputFiles = glob.glob("./*simtel.gz")

    # check the file size --- could be done by telescope type
    for afile in outputFiles:
        sizekb = os.path.getsize(afile) / 1024.0
        if sizekb < minSize:
            DIRAC.gLogger.warn(
                f"{afile}\n File size too small : {sizekb} < {minSize} kb"
            )
            DIRAC.gLogger.warn(f"Remove local File {afile}")
            os.remove(afile)
            outputFiles.remove(afile)  # remove from list of files processed
    DIRAC.gLogger.notice("Good files found:\n%s" % "\n".join(outputFiles))


def verifyGeneric(nbFiles=1, minSize=50.0, path="Data/*"):
    """Verify a PROD3 generic step

    Keyword arguments:
    nbFiles -- number of output files expected
    minSize -- minimum file size
    """
    DIRAC.gLogger.notice("Verifying generic step output")

    # get list of output files
    outputFiles = glob.glob(path)

    # check the number of output files
    N = len(outputFiles)
    if N != nbFiles:
        DIRAC.gLogger.error(f"Wrong number of output files : {N} instead of {nbFiles}")
        clean_output_file(outputFiles)
        DIRAC.exit(-1)

    # check the file size
    for afile in outputFiles:
        sizekb = os.path.getsize(afile) / 1024.0
        if sizekb < minSize:
            DIRAC.gLogger.error(
                f"{afile}\n File size too small : {sizekb} < {minSize} kb"
            )
            clean_output_file(outputFiles)
            DIRAC.exit(-1)
    DIRAC.gLogger.notice("Good files found:\n%s" % "\n".join(outputFiles))


@Script()
def main():
    """simple wrapper to put and register all PROD3 files

    Keyword arguments:
    args -- a list of arguments in order []
    """
    # check command line
    args = Script.getPositionalArgs()
    if len(args) < 1:
        DIRAC.gLogger.error("requires at least a step type")
        DIRAC.exit(-1)
    elif len(args) == 1:
        stepType = args[0]
    elif len(args) > 3:
        # now do something
        stepType = args[0]
        nbFiles = int(args[1])
        fileSize = float(args[2])
        if len(args) == 4:
            path = args[3]

    # What shall we verify ?
    if stepType == "corsika":
        verify_corsika()
    elif stepType == "simtel":
        verifySimtel(nbFiles, fileSize)
    elif stepType == "merging":
        verifyMerging(nbFiles, fileSize)
    elif stepType == "analysisinputs":
        verifyAnalysisInputs(fileSize)
    elif stepType == "generic":
        verifyGeneric(nbFiles, fileSize, path)
    else:
        DIRAC.gLogger.error(f'Do not know how to verify "{stepType}"')
        DIRAC.exit(-1)
    DIRAC.exit()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
  Add files to an existing transformation

  Usage:
    cta-transformation-add-files <transID> <ascii file with lfn list>
"""

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Base.Script import Script
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from CTADIRAC.Core.Utilities.tool_box import read_inputs_from_file


@Script()
def main():
    Script.parseCommandLine()
    args = Script.getPositionalArgs()
    if len(args) != 2:
        Script.showHelp()

    # get arguments
    transID = args[0]
    infile = args[1]
    infileList = read_inputs_from_file(infile)

    tc = TransformationClient()
    res = tc.addFilesToTransformation(transID, infileList)  # Files added here

    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)
    else:
        DIRAC.gLogger.notice(
            f"Successfully added {len(infileList)} files to transformation {transID}"
        )
        DIRAC.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Dump in a file the list of files for a given dataset

Usage:
   cta-prod-dump-dataset <datasetName or ascii file with a list of datasets>
"""

__RCSID__ = "$Id$"

import os
import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from CTADIRAC.Core.Utilities.tool_box import read_inputs_from_file
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=True)
    argss = Script.getPositionalArgs()
    fc = FileCatalogClient()

    datasetList = []
    if len(argss) == 1:
        if os.path.isfile(argss[0]):
            gLogger.notice(f"Reading datasets from input file: {argss[0]}")
            datasetList = read_inputs_from_file(argss[0])
        else:
            datasetList.append(argss[0])
    else:
        Script.showHelp()

    for dataset in datasetList:
        res = fc.getDatasetFiles(dataset)
        if not res["OK"]:
            gLogger.error("Failed to get files for dataset:", res["Message"])
            DIRAC.exit(-1)
        else:
            lfnList = res["Value"]["Successful"][dataset]
            f = open(dataset + ".list", "w")
            for lfn in lfnList:
                f.write(lfn + "\n")
            f.close()
            gLogger.notice(f"{len(lfnList)} files have been put in {dataset}.list")
    DIRAC.exit()


####################################################
if __name__ == "__main__":
    main()



""" Utility for building job tests
"""

import os
import json

from DIRAC import rootPath
from DIRAC.Interfaces.API.Job import Job
from DIRAC.Core.Workflow.Parameter import Parameter
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

from DIRAC.tests.Utilities.utils import find_all
from testJobUtils import base_to_all_jobs, end_of_all_jobs

dm = DataManager()
dirac = Dirac()
test_dir = "/vo.cta.in2p3.fr/tests/prodsys"
git_repo = "https://gitlab.cta-observatory.org/arrabito/mandel4ts.git"
workflow_path = "DIRAC/tests/Workflow"
exec_script = "exe-script.py"
mp_test = "mpTest.py"
mandelbrot_script = "./mandel4ts/mandelbrot.py"
mandelbrot_arg = "-P 0.0005 -M 1000 -L @{JOB_ID} -N 200"


def hello_world():
    """simple hello world job"""

    J = base_to_all_jobs("hello_world")
    try:
        J.setInputSandbox([find_all(exec_script, rootPath, workflow_path)[0]])
    except IndexError:
        J.setInputSandbox([find_all(exec_script, ".", workflow_path)[0]])
    J.setExecutable(exec_script, "", "hello_world.log")
    return end_of_all_jobs(J)


def hello_world_site_spec(site="CTAO.PIC.es"):
    """simple hello world job to specific site"""

    J = base_to_all_jobs(f"hello_world{site.split('.')[1]}")
    try:
        J.setInputSandbox([find_all(exec_script, rootPath, workflow_path)[0]])
    except IndexError:
        J.setInputSandbox([find_all(exec_script, ".", workflow_path)[0]])
    J.setExecutable(exec_script, "", "hello_world.log")
    J.setDestination(site)
    return end_of_all_jobs(J)


def mp_job(site="CTAO.PIC.es"):
    """simple hello world job with 6 CPUs"""

    J = base_to_all_jobs("mp_job")
    try:
        J.setInputSandbox([find_all(mp_test, rootPath, "DIRAC/tests/Utilities")[0]])
    except IndexError:
        J.setInputSandbox([find_all(mp_test, ".", "DIRAC/tests/Utilities")[0]])
    J.setExecutable(mp_test, "", "mp_job.log")
    J.setNumberOfProcessors(minNumberOfProcessors=6)
    J.setDestination(site)
    return end_of_all_jobs(J)


def hello_world_input(input_lfn):
    """simple hello world job with input data"""
    J = base_to_all_jobs("hello_worldInput")
    J.setInputData(input_lfn)
    J.setExecutable("ls", "-ltrA", "hello_worldInput.log")
    return end_of_all_jobs(J)


def hello_world_output(output_file, exe_script, se):
    """Job with Output Data"""
    J = base_to_all_jobs("hello_worldOutput")
    with open(exe_script, "w") as script:
        script.write("#!/bin/bash\n")
        script.write(f"echo 'This is a test file' > {output_file}\n")
    J.setExecutable(exe_script)
    J.setExecutable("ls", arguments="-ltrA")
    J.setOutputData(output_file, output_se=se)
    return end_of_all_jobs(J)


def mandelbrot_simulation():
    J = base_to_all_jobs("mandelbrot_simulation")
    J.setExecutable(f"git clone {git_repo}")
    J.setExecutable(mandelbrot_script, arguments=mandelbrot_arg)
    return end_of_all_jobs(J)


def create_workflow_body_step():
    job = Job()
    job.setName("mandelbrot raw")
    job.setOutputSandbox(["*log"])
    # this is so that the JOB_ID within the transformation can be evaluated on the fly in the job application, see below
    job.workflow.addParameter(
        Parameter(
            "JOB_ID", "000000", "string", "", "", True, False, "Initialize JOB_ID"
        )
    )
    # define the job workflow in 3 steps
    # job step1: setup software
    job.setExecutable(f"git clone {git_repo}")
    # job step2: run mandelbrot application
    # note how the JOB_ID (within the transformation) is passed as an argument and will be evaluated on the fly
    job.setExecutable(mandelbrot_script, arguments=mandelbrot_arg)
    return job.workflow.toXML()


def create_workflow_body_step1(output_se):
    job = Job()
    job.setName("mandelbrot raw")
    job.setOutputSandbox(["*log"])
    # this is so that the JOB_ID within the transformation can be evaluated on the fly in the job application, see below
    job.workflow.addParameter(
        Parameter(
            "JOB_ID", "000000", "string", "", "", True, False, "Initialize JOB_ID"
        )
    )
    # define the job workflow in 3 steps
    # job step1: setup software
    job.setExecutable(f"git clone {git_repo}")
    # job step2: run mandelbrot application
    # note how the JOB_ID (within the transformation) is passed as an argument and will be evaluated on the fly
    job.setExecutable(mandelbrot_script, arguments=mandelbrot_arg)
    # job step3: upload data and set metadata
    output_path = os.path.join(test_dir, "mandelbrot/images/raw")
    output_pattern = "data_*txt"
    output_metadata = json.dumps(
        {
            "application": "mandelbrot",
            "image_format": "ascii",
            "image_width": 7680,
            "image_height": 200,
        }
    )
    job.setExecutable(
        "./mandel4ts/dirac-add-files.py",
        arguments=f"{output_path} '{output_pattern}' {output_se} '{output_metadata}'",
    )
    return job.workflow.toXML()


def create_workflow_body_step2(output_se):
    job = Job()
    job.setName("merge mandelbrot")
    job.setOutputSandbox(["*log"])
    # define the job workflow in 3 steps
    # job step1: setup software
    job.setExecutable(f"git clone {git_repo}")
    # job step2: run mandelbrot merge
    job.setExecutable("./mandel4ts/merge_data.py")
    # job step3: upload data and set metadata
    output_path = os.path.join(test_dir, "mandelbrot/images/merged")
    output_pattern = "data_merged*txt"
    nb_input_files = 7
    output_metadata = json.dumps(
        {
            "application": "mandelbrot",
            "image_format": "ascii",
            "image_width": 7680,
            "image_height": 200 * nb_input_files,
        }
    )
    job.setExecutable(
        "./mandel4ts/dirac-add-files.py",
        arguments=f"{output_path} '{output_pattern}' {output_se} '{output_metadata}'",
    )
    return job.workflow.toXML()


def clean_dir(dir):
    """Clean directory"""
    res = dm.cleanLogicalDirectory(dir)
    return res


def add_file(lfn, se):
    """Upload file to SE"""
    res = dirac.addFile(lfn, os.path.basename(lfn), se)
    return res



from DIRAC.TransformationSystem.Service.TransformationManagerHandler import (
    TransformationManagerHandler as TManagerBase,
)


class TransformationManagerHandlerMixin:
    types_get_tasks_by_file_status = [int, str]

    @classmethod
    def export_get_tasks_by_file_status(self, trans_id, file_status):
        return self.transformationDB.get_tasks_by_file_status(trans_id, file_status)


class TransformationManagerHandler(TransformationManagerHandlerMixin, TManagerBase):
    pass
